@@ -1,10 +1,10 @@
 /*
- * Author: Paul.Russell@rustcorp.com.au
+ * Author: Paul ``Rusty'' Russell.  ipchains@rustcorp.com
  *
  * Based on the ipfwadm code by Jos Vos <jos@xos.nl> (see README).
  *
  *	ipchains -- IP firewall administration for kernels with
- * CONFIG_IP_FIREWALL_CHAINS.
+ * firewall chains (mainly 2.1.102 and above).
  *
  *	See the accompanying manual page ipchains(8) for information
  *	about proper usage of this program.
@@ -75,8 +75,24 @@
  *        Fixed libipfwc bug with ipfwc_check_packet().
  * 1.3.8: -L now works for listing chains other than input.
  *        [ Thanks to Bernhard Weisshuhn ].
+ * 1.3.9: Added --line-numbers
+ *        [ Thanks to Danek Duvall ].
+ *        Fixed --delete-chains (not --delete) in help.
+ *        libipfwc now gives "" target for acct rules, so we turn it into
+ *        "-" in print_firewall.
+ *        Fixed parsing of `--sport ! 53'; same as `! --sport 53'.
+ *	  Added enabling forward warning based on Andrew Wansink's patch.
+ * 1.3.10: --delete-chain now takes optional arg, like -X.
+ *        [ Found by Lothar Gerlach ]
+ *        Maximize-throughput not Minimize-throughput for TOS!
+ *        [ Found by Adam Kumiszcza ]
+ *        Warn about `-i !eth0' and `-i eth0*' cases.
+ *        [ Thanks to John Martinez ]
+ *        ICMP numbers printed in -h icmp
+ *        [ Thanks to Brett Eldridge ]
+ *        Fixed wildcard interfaces getting extra + with `-A input -v'.
+ *        [ Thanks to Howard Lowndes ]
  */
-
 #include <stdio.h>
 #include <stdarg.h>
 #include <errno.h>
@@ -90,6 +106,7 @@
 #include <sys/param.h>
 #include <pwd.h>
 #include <sys/types.h>
+#include <termios.h>
 
 #include "libipfwc/libipfwc.h"
 
@@ -153,9 +170,10 @@ static const char cmdflags[] = { 'I', 'D', 'D', 'R', 'A', 'L', 'F', 'Z',
 #define OPT_SRCPT       0x04000U
 #define OPT_DSTPT       0x08000U
 #define OPT_NETLINK     0x10000U
-#define NUMBER_OF_OPT	17
+#define OPT_LINENUMBERS 0x20000U
+#define NUMBER_OF_OPT	18
 static const char optflags[] = { 'n', 's', 'd', 'p', 'j', 'y', 'b', 'v',
-				 'l', 'x', 't', 'i', 'f', 'm', 's', 'd', 'o'};
+				 'l', 'x', 't', 'i', 'f', 'm', 's', 'd', 'o', '3'};
 
 #define FMT_NUMERIC	0x0001
 #define FMT_NOCOUNTS	0x0002
@@ -170,6 +188,7 @@ static const char optflags[] = { 'n', 's', 'd', 'p', 'j', 'y', 'b', 'v',
 #define FMT_TOS		0x0400
 #define FMT_MARK        0x0800
 #define FMT_NETLINK     0x1000
+#define FMT_LINENUMBERS 0x2000
 
 #define FMT_PRINT_RULE (FMT_NOCOUNTS | FMT_OPTIONS | FMT_TOS | FMT_VIA \
 			| FMT_NUMERIC | FMT_NOTABLE | FMT_MARK | FMT_NETLINK)
@@ -184,7 +203,7 @@ struct option opts[] = {
 	{ "zero", 2, 0,  'Z' },
 	{ "check", 1, 0,  'C' },
 	{ "new-chain", 1, 0,  'N' },
-	{ "delete-chain", 1, 0,  'X' },
+	{ "delete-chain", 2, 0,  'X' },
 	{ "policy", 1, 0,  'P' },
 	{ "masquerading", 0, 0,  'M' },
 	{ "set", 1, 0,  'S' },
@@ -198,7 +217,7 @@ struct option opts[] = {
 	{ "icmp-type", 1, 0, '1' },
 	{ "sport", 1, 0, '1' }, /* synonym */
 	{ "dport", 1, 0, '2' }, /* synonym */
-	{ "proto", 1, 0,  'p' },
+	{ "protocol", 1, 0,  'p' },
 	{ "interface", 1, 0, 'i' },
 	{ "jump", 1, 0, 'j' },
 	{ "TOS", 1, 0, 't' },
@@ -211,7 +230,9 @@ struct option opts[] = {
 	{ "fragments", 0, 0, 'f' },
 	{ "syn", 0, 0, 'y' },
 	{ "version", 0, 0, 'V' },
-	{ "help", 0, 0, 'h'},
+	{ "help", 0, 0, 'h' },
+	{ "line-numbers", 0, 0, '3' },
+	{ "no-warnings", 0, 0, 'W' },
 	{0}
 };
 
@@ -223,14 +244,14 @@ struct TOS_value
 } TOS_values[] = {
 #ifdef NOT_A_GODDAMN_YANK
 	{ 0x10, "Minimise Delay" },
-	{ 0x08, "Minimise Throughput" },
+	{ 0x08, "Maximise Throughput" },
 	{ 0x04, "Maximise Reliability" },
-	{ 0x01, "Minimise Cost" },
+	{ 0x02, "Minimise Cost" },
 #else
 	{ 0x10, "Minimize Delay" },
-	{ 0x08, "Minimize Throughput" },
+	{ 0x08, "Maximize Throughput" },
 	{ 0x04, "Maximize Reliability" },
-	{ 0x01, "Minimize Cost" },
+	{ 0x02, "Minimize Cost" },
 #endif
 	{ 0x00, "Normal Service" },
 };
@@ -247,21 +268,21 @@ struct TOS_value
 static char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
 /* Well, it's better than "Re: Linux vs FreeBSD" */
 {
-	/*     -n  -s  -d  -p  -j  -y  -b  -v  -l  -x  -t  -i  -f  -m  -s  -d  -o */
-/*INSERT*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' '},
-/*DELETE*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' '},
-/*DELETE_NUM*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*REPLACE*/   {'x',' ',' ',' ',' ',' ','x',' ',' ','x',' ',' ',' ',' ',' ',' ',' '},
-/*APPEND*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' '},
-/*LIST*/      {' ','x','x','x','x','x','x',' ','x',' ','x','x','x','x','x','x','x'},
-/*FLUSH*/     {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*ZERO*/      {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*NEW_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*DEL_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*SET_POLICY*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*MASQUERADE*/{'x','x','x','x','x','x','x','x','x','x','x','x','x','x','x','x','x'},
-/*SET_MASQ*/  {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x'},
-/*CHECK*/     {'x','+','+','+','x',' ',' ',' ','x','x','x','+',' ','x',' ',' ','x'}
+	/*     -n  -s  -d  -p  -j  -y  -b  -v  -l  -x  -t  -i  -f  -m  -s  -d  -o  -3*/
+/*INSERT*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
+/*DELETE*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
+/*DELETE_NUM*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*REPLACE*/   {'x',' ',' ',' ',' ',' ','x',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
+/*APPEND*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
+/*LIST*/      {' ','x','x','x','x','x','x',' ','x',' ','x','x','x','x','x','x','x',' '},
+/*FLUSH*/     {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*ZERO*/      {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*NEW_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*DEL_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*SET_POLICY*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*MASQUERADE*/{'x','x','x','x','x','x','x','x','x','x','x','x','x','x','x','x','x','x'},
+/*SET_MASQ*/  {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
+/*CHECK*/     {'x','+','+','+','x',' ',' ',' ','x','x','x','+',' ','x',' ',' ','x','x'}
 };
 
 static int inverse_for_options[NUMBER_OF_OPT] =
@@ -302,13 +323,10 @@ struct masq_timeout {
 	int	udp_timeout;
 } timeouts;
 
-static char package_version[] = "ipchains 1.3.8, 27-Oct-1998";
+static char package_version[] = "ipchains 1.3.10, 1-Sep-2000";
 
 static const char *program;
 
-/* Do we invert the next option? */
-static int invert = FALSE;
-
 static int check_inverse(const char option[], int *invert);
 
 /* Parsing functions for command line. */
@@ -325,7 +343,8 @@ static void parse_ports(const char *portstring, __u16 *ports, __u16 proto);
 static __u16 parse_port(const char *port, unsigned short proto);
 static unsigned short parse_mark(char *markstring, __u32 *mark);
 static unsigned short parse_outputsize(char *outputstring, __u16 *outputsize);
-static unsigned short parse_interface(const char *ifstring, char *vianame);
+static unsigned short parse_interface(const char *ifstring, char *vianame,
+				      int warnings);
 static void parse_hexbyte(char *, unsigned char *);
 static int parse_timeout(char *);
 static const char *parse_target(const char *targetname);
@@ -333,13 +352,13 @@ static int parse_rulenumber(const char *rule);
 
 /* Various parsing helper functions */
 static struct in_addr *host_to_addr(const char *, unsigned int *);
-static char *addr_to_host(struct in_addr *);
+static char *addr_to_host(const struct in_addr *);
 static struct in_addr *network_to_addr(const char *);
-static char *addr_to_network(struct in_addr *);
-static char *addr_to_anyname(struct in_addr *);
+static char *addr_to_network(const struct in_addr *);
+static char *addr_to_anyname(const struct in_addr *);
 static struct in_addr *dotted_to_addr(const char *);
-static char *addr_to_dotted(struct in_addr *);
-static char *mask_to_dotted(struct in_addr *);
+static char *addr_to_dotted(const struct in_addr *);
+static char *mask_to_dotted(const struct in_addr *);
 static int service_to_port(const char *, unsigned short);
 static char *port_to_service(int, unsigned short);
 static int string_to_number(const char *, int, int);
@@ -348,20 +367,25 @@ static int string_to_number(const char *, int, int);
 static void generic_opt_check(int command, int options);
 static char opt2char(int option);
 static char cmd2char(int option);
-static void add_command(int *cmd, const int newcmd, const int othercmds);
+static void add_command(int *cmd, const int newcmd, const int othercmds,
+			int *invert);
 
 /* Check for illegal TOS manipulation */
-static void check_tos(unsigned char tosand, unsigned char tosxor);
+static void check_tos(unsigned char tosand, unsigned char tosxor,
+		      int warnings);
 
 static int list_masq(unsigned int options);
 
-static void print_firewall(FILE *, struct ip_fwuser *, __u64, __u64, int);
+static void print_firewall(FILE *, const struct ip_fwuser *,
+			   int, __u64, __u64, int);
 static void print_masq(FILE *, struct masq *, int);
 static int read_masqinfo(FILE *, struct masq *, int);
+static void warn_ipforward(void);
 
 static void set_option(unsigned int *options,
 		       unsigned int option,
-		       __u16 *invflg);
+		       __u16 *invflg,
+		       int *invert);
 static void inaddrcpy(struct in_addr *, struct in_addr *);
 static void *fw_malloc(size_t);
 static void *fw_calloc(size_t, size_t);
@@ -379,6 +403,7 @@ static void exit_printhelp() __attribute__((noreturn));
 
 
 struct icmp_names {
+	const char *columns;
 	const char *name;
 	__u16 type;
 	__u16 code_min, code_max;
@@ -386,57 +411,56 @@ struct icmp_names {
 };
 
 const struct icmp_names icmp_codes[] = {
-	{ "echo-reply", 0, 0, 0xFFFF, TRUE },
-	/* Alias */ { "pong", 0, 0, 0xFFFF, TRUE },
+	{ "0    0    ", "echo-reply", 0, 0, 0xFFFF, TRUE },
+	{ "", "pong", 0, 0, 0xFFFF, TRUE },
+	{ "3         ", "destination-unreachable", 3, 0, 0xFFFF, FALSE },
+	{ "     0      ", "network-unreachable", 3, 0, 0, TRUE },
+	{ "     1      ", "host-unreachable", 3, 1, 1, TRUE },
+	{ "     2      ", "protocol-unreachable", 3, 2, 2, TRUE },
+	{ "     3      ", "port-unreachable", 3, 3, 3, TRUE },
+	{ "     4      ", "fragmentation-needed", 3, 4, 4, TRUE },
+	{ "     5      ", "source-route-failed", 3, 5, 5, TRUE },
+	{ "     6      ", "network-unknown", 3, 6, 6, TRUE },
+	{ "     7      ", "host-unknown", 3, 7, 7, TRUE },
+	{ "     9      ", "network-prohibited", 3, 9, 9, TRUE },
+	{ "     10     ", "host-prohibited", 3, 10, 10, TRUE },
+	{ "     11     ", "TOS-network-unreachable", 3, 11, 11, TRUE },
+	{ "     12     ", "TOS-host-unreachable", 3, 12, 12, TRUE },
+	{ "     13     ", "communication-prohibited", 3, 13, 13, TRUE },
+	{ "     14     ", "host-precedence-violation", 3, 14, 14, TRUE },
+	{ "     15     ", "precedence-cutoff", 3, 15, 15, TRUE },
 
-	{ "destination-unreachable", 3, 0, 0xFFFF, FALSE },
-	{   "network-unreachable", 3, 0, 0, TRUE },
-	{   "host-unreachable", 3, 1, 1, TRUE },
-	{   "protocol-unreachable", 3, 2, 2, TRUE },
-	{   "port-unreachable", 3, 3, 3, TRUE },
-	{   "fragmentation-needed", 3, 4, 4, TRUE },
-	{   "source-route-failed", 3, 5, 5, TRUE },
-	{   "network-unknown", 3, 6, 6, TRUE },
-	{   "host-unknown", 3, 7, 7, TRUE },
-	{   "network-prohibited", 3, 9, 9, TRUE },
-	{   "host-prohibited", 3, 10, 10, TRUE },
-	{   "TOS-network-unreachable", 3, 11, 11, TRUE },
-	{   "TOS-host-unreachable", 3, 12, 12, TRUE },
-	{   "communication-prohibited", 3, 13, 13, TRUE },
-	{   "host-precedence-violation", 3, 14, 14, TRUE },
-	{   "precedence-cutoff", 3, 15, 15, TRUE },
+	{ "4    0    ", "source-quench", 4, 0, 0xFFFF, TRUE },
 
-	{ "source-quench", 4, 0, 0xFFFF, TRUE },
+	{ "5         ", "redirect", 5, 0, 0xFFFF, FALSE },
+	{ "     0      ", "network-redirect", 5, 0, 0, TRUE },
+	{ "     1      ", "host-redirect", 5, 1, 1, TRUE },
+	{ "     2      ", "TOS-network-redirect", 5, 2, 2, TRUE },
+	{ "     3      ", "TOS-host-redirect", 5, 3, 3, TRUE },
 
-	{ "redirect", 5, 0, 0xFFFF, FALSE },
-	{   "network-redirect", 5, 0, 0, TRUE },
-	{   "host-redirect", 5, 1, 1, TRUE },
-	{   "TOS-network-redirect", 5, 2, 2, TRUE },
-	{   "TOS-host-redirect", 5, 3, 3, TRUE },
+	{ "8    0    ", "echo-request", 8, 0, 0xFFFF, TRUE },
+	{ "", "ping", 8, 0, 0xFFFF, TRUE },
 
-	{ "echo-request", 8, 0, 0xFFFF, TRUE },
-	/* Alias */ { "ping", 8, 0, 0xFFFF, TRUE },
+	{ "9    0    ", "router-advertisement", 9, 0, 0xFFFF, TRUE },
+	
+	{ "10   0    ", "router-solicitation", 10, 0, 0xFFFF, TRUE }, 
 
-	{ "router-advertisement", 9, 0, 0xFFFF, TRUE },
+	{ "11        ", "time-exceeded", 11, 0, 0xFFFF, FALSE }, 
+	{ "", "ttl-exceeded", 11, 0, 0xFFFF, FALSE },
+	{ "     0      ", "ttl-zero-during-transit", 11, 0, 0, TRUE },
+	{ "     1      ", "ttl-zero-during-reassembly", 11, 1, 1, TRUE },
 
-	{ "router-solicitation", 10, 0, 0xFFFF, TRUE },
+	{ "12        ", "parameter-problem", 12, 0, 0xFFFF, FALSE },
+	{ "     0      ", "ip-header-bad", 12, 0, 0, TRUE },
+	{ "     1      ", "required-option-missing", 12, 1, 1, TRUE },
 
-	{ "time-exceeded", 11, 0, 0xFFFF, FALSE },
-	/* Alias */ { "ttl-exceeded", 11, 0, 0xFFFF, FALSE },
-	{   "ttl-zero-during-transit", 11, 0, 0, TRUE },
-	{   "ttl-zero-during-reassembly", 11, 1, 1, TRUE },
+	{ "13   0    ", "timestamp-request", 13, 0, 0xFFFF, TRUE },
 
-	{ "parameter-problem", 12, 0, 0xFFFF, FALSE },
-	{   "ip-header-bad", 12, 0, 0, TRUE },
-	{   "required-option-missing", 12, 1, 1, TRUE },
+	{ "14   0    ", "timestamp-reply", 14, 0, 0xFFFF, TRUE },
 
-	{ "timestamp-request", 13, 0, 0xFFFF, TRUE },
+	{ "17   0    ", "address-mask-request", 17, 0, 0xFFFF, TRUE },
 
-	{ "timestamp-reply", 14, 0, 0xFFFF, TRUE },
-
-	{ "address-mask-request", 17, 0, 0xFFFF, TRUE },
-
-	{ "address-mask-reply", 18, 0, 0xFFFF, TRUE }
+	{ "18   0    ", "address-mask-reply", 18, 0, 0xFFFF, TRUE }
 };
 
 static void
@@ -494,7 +518,7 @@ append_entry(const ip_chainlabel chain,
 		for (j = 0; j < ndaddrs; j++) {
 			ipfw.ipfw.fw_dst.s_addr = daddrs[j].s_addr;
 			if (verbose)
-				print_firewall(stdout, &ipfw,
+				print_firewall(stdout, &ipfw, 0,
 					       0, 0, FMT_PRINT_RULE);
 			ret &= ipfwc_append_entry(chain, &ipfw);
 		}
@@ -521,7 +545,7 @@ replace_entry(const ip_chainlabel chain,
 	ipfw.ipfw.fw_dst.s_addr = daddr->s_addr;
 
 	if (verbose)
-		print_firewall(stdout, &ipfw, 0, 0, FMT_PRINT_RULE);
+		print_firewall(stdout, &ipfw, rulenum, 0, 0, FMT_PRINT_RULE);
 	return ipfwc_replace_entry(chain, &ipfw, rulenum);
 }
 
@@ -544,7 +568,7 @@ insert_entry(const ip_chainlabel chain,
 		for (j = 0; j < ndaddrs; j++) {
 			ipfw.ipfw.fw_dst.s_addr = daddrs[j].s_addr;
 			if (verbose)
-				print_firewall(stdout, &ipfw,
+				print_firewall(stdout, &ipfw, rulenum,
 					       0, 0, FMT_PRINT_RULE);
 			ret &= ipfwc_insert_entry(chain, &ipfw, rulenum);
 		}
@@ -576,7 +600,7 @@ delete_entry(const ip_chainlabel chain,
 		for (j = 0; j < ndaddrs; j++) {
 			ipfw.ipfw.fw_dst.s_addr = daddrs[j].s_addr;
 			if (verbose)
-				print_firewall(stdout, &ipfw,
+				print_firewall(stdout, &ipfw, 0,
 					       0, 0, FMT_PRINT_RULE);
 			ret &= ipfwc_delete_entry(chain, &ipfw);
 		}
@@ -609,7 +633,7 @@ check_packet(const ip_chainlabel chain,
 		for (j = 0; j < ndaddrs; j++) {
 			ipfw.ipfw.fw_dst.s_addr = daddrs[j].s_addr;
 			if (verbose)
-				print_firewall(stdout, &ipfw, 0, 0,
+				print_firewall(stdout, &ipfw, 0, 0, 0,
 					       FMT_PRINT_RULE);
 			msg = ipfwc_check_packet(chain, &ipfw.ipfw);
 			if (!msg) ret = 0;
@@ -685,10 +709,10 @@ delete_chain(const ip_chainlabel chain, int verbose)
 
 static int
 list_entries(const ip_chainlabel chain, int zero, int verbose, int numeric,
-	     int expanded)
+	     int expanded, int linenumbers)
 {
 	int format, found = 0;
-	unsigned int num_rules, num_chains, i;
+	unsigned int num_rules, num_chains, i, num;
 	struct ipfwc_fwchain *chains = ipfwc_get_chainnames(&num_chains);
 	struct ipfwc_fwrule *rules;
 
@@ -713,6 +737,9 @@ list_entries(const ip_chainlabel chain, int zero, int verbose, int numeric,
 	if (!expanded)
 		format |= FMT_KILOMEGAGIGA;
 
+	if (linenumbers)
+		format |= FMT_LINENUMBERS;
+
 	/* This is unreliable if a new chain is created and a rule
            added between ipfwc_get_chainnames and ipfwc_get_rules.
 	*/
@@ -737,11 +764,13 @@ list_entries(const ip_chainlabel chain, int zero, int verbose, int numeric,
 			printf("\n");
 			/* First one gets header. */
 			format |= FMT_HEADER;
+			num = 1;
 			for (rulenum = 0; rulenum < num_rules; rulenum++) {
 				if (strcmp(rules[rulenum].chain->label, 
 					   chains[i].label) == 0) {
 					print_firewall(stdout, 
 						       &rules[rulenum].ipfw,
+						       num++,
 						       rules[rulenum].bytes,
 						       rules[rulenum].packets,
 						       format);
@@ -773,6 +802,9 @@ main(int argc, char *argv[])
 	const char *policy = NULL;
 	unsigned int rulenum = 0, options = 0, command = 0;
 	int ret = 1;
+	int warnings = 1;
+	/* Do we invert the next option? */
+	int invert = FALSE;
 
 	program = argv[0];
 
@@ -784,12 +816,12 @@ main(int argc, char *argv[])
 			 * Command selection
 			 */
 		case 'A':
-			add_command(&command, CMD_APPEND, CMD_NONE);
+			add_command(&command, CMD_APPEND, CMD_NONE, &invert);
 			chain = optarg;
 			break;
 
 		case 'D':
-			add_command(&command, CMD_DELETE, CMD_NONE);
+			add_command(&command, CMD_DELETE, CMD_NONE, &invert);
 			chain = optarg;
 			if (optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!') {
@@ -799,12 +831,12 @@ main(int argc, char *argv[])
 			break;
 
 		case 'C':
-			add_command(&command, CMD_CHECK, CMD_NONE);
+			add_command(&command, CMD_CHECK, CMD_NONE, &invert);
 			chain = optarg;
 			break;
 
 		case 'R':
-			add_command(&command, CMD_REPLACE, CMD_NONE);
+			add_command(&command, CMD_REPLACE, CMD_NONE, &invert);
 			chain = optarg;
 			if (optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!') {
@@ -817,7 +849,7 @@ main(int argc, char *argv[])
 			break;
 
 		case 'I':
-			add_command(&command, CMD_INSERT, CMD_NONE);
+			add_command(&command, CMD_INSERT, CMD_NONE, &invert);
 			chain = optarg;
 			if (optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!') {
@@ -828,7 +860,7 @@ main(int argc, char *argv[])
 
 		case 'L':
 			add_command(&command, CMD_LIST,
-				    CMD_ZERO|CMD_MASQUERADE);
+				    CMD_ZERO|CMD_MASQUERADE, &invert);
 			if (optarg) chain = optarg;
 			else if (optind < argc && argv[optind][0] != '-'
 				 && argv[optind][0] != '!')
@@ -843,7 +875,7 @@ main(int argc, char *argv[])
 			break;
 
 		case 'F':
-			add_command(&command, CMD_FLUSH, CMD_NONE);
+			add_command(&command, CMD_FLUSH, CMD_NONE, &invert);
 			if (optarg) chain = optarg;
 			else if (optind < argc && argv[optind][0] != '-'
 				 && argv[optind][0] != '!')
@@ -851,7 +883,7 @@ main(int argc, char *argv[])
 			break;
 
 		case 'Z':
-			add_command(&command, CMD_ZERO, CMD_LIST);
+			add_command(&command, CMD_ZERO, CMD_LIST, &invert);
 			if (optarg) chain = optarg;
 			else if (optind < argc && argv[optind][0] != '-'
 				&& argv[optind][0] != '!')
@@ -866,12 +898,14 @@ main(int argc, char *argv[])
 			break;
 
 		case 'N':
-			add_command(&command, CMD_NEW_CHAIN, CMD_NONE);
+			add_command(&command, CMD_NEW_CHAIN, CMD_NONE,
+				    &invert);
 			chain = optarg;
 			break;
 
 		case 'X':
-			add_command(&command, CMD_DELETE_CHAIN, CMD_NONE);
+			add_command(&command, CMD_DELETE_CHAIN, CMD_NONE,
+				    &invert);
 			if (optarg) chain = optarg;
 			else if (optind < argc && argv[optind][0] != '-'
 				 && argv[optind][0] != '!')
@@ -879,7 +913,8 @@ main(int argc, char *argv[])
 			break;
 
 		case 'P':
-			add_command(&command, CMD_SET_POLICY, CMD_NONE);
+			add_command(&command, CMD_SET_POLICY, CMD_NONE,
+				    &invert);
 			chain = optarg;
 			if (optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!') {
@@ -893,13 +928,13 @@ main(int argc, char *argv[])
 
 		case 'M':
 			add_command(&command, CMD_MASQUERADE,
-				    CMD_LIST|CMD_SET_MASQUERADE);
+				    CMD_LIST|CMD_SET_MASQUERADE, &invert);
 			break;
 
 		case 'S':
 #ifndef	IPFWADM_NO_TIMEOUT
 			add_command(&command, CMD_SET_MASQUERADE,
-				    CMD_MASQUERADE);
+				    CMD_MASQUERADE, &invert);
 			if (optind + 1 < argc
 			    && argv[optind][0] != '!'
 			    && argv[optind][0] != '-'
@@ -935,7 +970,8 @@ main(int argc, char *argv[])
 		case 'p':
 			if (check_inverse(optarg, &invert))
 				optind++;
-			set_option(&options, OPT_PROTOCOL, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_PROTOCOL, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_proto = parse_protocol(argv[optind-1]);
 			if (fw.ipfw.fw_proto == 0
 			    && (fw.ipfw.fw_invflg & IP_FW_INV_PROTO))
@@ -946,17 +982,21 @@ main(int argc, char *argv[])
 		case 's':
 			if (check_inverse(optarg, &invert))
 				optind++;
-			set_option(&options, OPT_SOURCE, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_SOURCE, &fw.ipfw.fw_invflg,
+				   &invert);
 
 			shostnetworkmask = argv[optind-1];
-			if (check_inverse(argv[optind], &invert))
-				optind++;
-			if (optind < argc && argv[optind][0] != '-'
-			    && argv[optind][0] != '!')
+			check_inverse(argv[optind], &invert);
+
+			if (optind+invert < argc
+			    && argv[optind+invert][0] != '-'
+			    && argv[optind+invert][0] != '!')
 			{
+				optind += invert;
 				srcpts = argv[optind++];
 				set_option(&options, OPT_SRCPT,
-					   &fw.ipfw.fw_invflg);
+					   &fw.ipfw.fw_invflg,
+					   &invert);
 			}
 			break;
 
@@ -964,39 +1004,43 @@ main(int argc, char *argv[])
 			if (check_inverse(optarg, &invert))
 				optind++;
 			set_option(&options, OPT_DESTINATION,
-				   &fw.ipfw.fw_invflg);
+				   &fw.ipfw.fw_invflg,
+				   &invert);
 
 			dhostnetworkmask = argv[optind-1];
-			if (check_inverse(argv[optind], &invert))
-				optind++;
-			if (optind < argc && argv[optind][0] != '-'
-			    && argv[optind][0] != '!')
+			check_inverse(argv[optind], &invert);
+
+			if (optind + invert < argc
+			    && argv[optind+invert][0] != '-'
+			    && argv[optind+invert][0] != '!')
 			{
+				optind += invert;
 				dstpts = argv[optind++];
 				set_option(&options, OPT_DSTPT,
-					   &fw.ipfw.fw_invflg);
+					   &fw.ipfw.fw_invflg,
+					   &invert);
 			}
 			break;
 
 		case '1':
 			if (check_inverse(optarg, &invert))
 				optind++;
-			set_option(&options, OPT_SRCPT, &fw.ipfw.fw_invflg);
-			srcpts = optarg;
+			set_option(&options, OPT_SRCPT, &fw.ipfw.fw_invflg,
+				   &invert);
+			srcpts = argv[optind-1];
 			break;
 
 		case '2':
 			if (check_inverse(optarg, &invert))
 				optind++;
-			set_option(&options, OPT_DSTPT, &fw.ipfw.fw_invflg);
-			dstpts = optarg;
+			set_option(&options, OPT_DSTPT, &fw.ipfw.fw_invflg,
+				   &invert);
+			dstpts = argv[optind-1];
 			break;
 
 		case 'j':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_JUMP, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_JUMP, &fw.ipfw.fw_invflg,
+				   &invert);
 			jumpto = parse_target(optarg);
 			if (strcmp(jumpto, IP_FW_LABEL_REDIRECT) == 0
 			    && optind < argc && argv[optind][0] != '-'
@@ -1007,59 +1051,49 @@ main(int argc, char *argv[])
 		case 'i':
 			if (check_inverse(optarg, &invert))
 				optind++;
-			set_option(&options, OPT_VIANAME, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_VIANAME, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_flg |= parse_interface(argv[optind-1],
-							  fw.ipfw.fw_vianame);
+							  fw.ipfw.fw_vianame,
+							  warnings);
 			break;
 
 		case 'f':
-			set_option(&options, OPT_FRAGMENT, &fw.ipfw.fw_invflg);
-			invert = FALSE;
+			set_option(&options, OPT_FRAGMENT, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_flg |= IP_FW_F_FRAG;
 			break;
 
 		case 'b':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_BIDIR, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_BIDIR, &fw.ipfw.fw_invflg,
+				   &invert);
 			break;
 
 		case 'v':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_VERBOSE, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_VERBOSE, &fw.ipfw.fw_invflg,
+				   &invert);
 			break;
 
 		case 'm':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_MARK, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_MARK, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_flg |= parse_mark(optarg, &fw.ipfw.fw_mark);
 			break;
 
 		case 'n':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_NUMERIC, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_NUMERIC, &fw.ipfw.fw_invflg,
+				   &invert);
 			break;
 
 		case 'l':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_PRINTK, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_PRINTK, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_flg |= IP_FW_F_PRN;
 			break;
 
 		case 'o':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_NETLINK, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_NETLINK, &fw.ipfw.fw_invflg,
+				   &invert);
 			if (!optarg && optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!')
 				optarg = argv[optind++];
@@ -1070,17 +1104,13 @@ main(int argc, char *argv[])
 			break;
 
 		case 't':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_TOS, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_TOS, &fw.ipfw.fw_invflg,
+				   &invert);
 			if (optind < argc && argv[optind][0] != '-'
 			    && argv[optind][0] != '!') {
 				parse_hexbyte(optarg, &fw.ipfw.fw_tosand);
 				parse_hexbyte(argv[optind++],
 					      &fw.ipfw.fw_tosxor);
-				check_tos(fw.ipfw.fw_tosand,
-					  fw.ipfw.fw_tosxor);
 			} else
 				exit_error(PARAMETER_PROBLEM,
 					   "-%c requires 2 hexbyte arguments",
@@ -1088,16 +1118,14 @@ main(int argc, char *argv[])
 			break;
 
 		case 'x':
-			if (invert)
-				exit_error(PARAMETER_PROBLEM,
-					   "unexpected ! flag");
-			set_option(&options, OPT_EXPANDED, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_EXPANDED, &fw.ipfw.fw_invflg,
+				   &invert);
 			break;
 
 		case 'y':
-			set_option(&options, OPT_TCPSYN, &fw.ipfw.fw_invflg);
+			set_option(&options, OPT_TCPSYN, &fw.ipfw.fw_invflg,
+				   &invert);
 			fw.ipfw.fw_flg |= IP_FW_F_TCPSYN;
-			invert = FALSE;
 			break;
 
 		case 'V':
@@ -1107,6 +1135,22 @@ main(int argc, char *argv[])
 				printf("%s\n", package_version);
 			exit(0);
 
+		case '3':
+			if (invert)
+				exit_error(PARAMETER_PROBLEM,
+					   "cannot have ! before --line-numbers");
+			set_option(&options, OPT_LINENUMBERS, 
+				   &fw.ipfw.fw_invflg,
+				   &invert);
+			break;
+
+		case 'W':
+			if (invert)
+				exit_error(PARAMETER_PROBLEM,
+					   "cannot have ! before --no-warnings");
+			warnings = 0;
+			break;
+
 		case 1: /* non option */
 			if (optarg[0] == '!' && optarg[1] == '\0') {
 				if (invert)
@@ -1227,6 +1271,12 @@ main(int argc, char *argv[])
 		exit_error(PARAMETER_PROBLEM, "Replacement rule does not "
 			   "specify a unique address");
 
+	if (warnings 
+	    && chain
+	    && strcmp(chain, IP_FW_LABEL_FORWARD) == 0
+	    && (command == CMD_APPEND || command == CMD_INSERT))
+		warn_ipforward();
+
 	/* Allow lazy -S without -M */
 	if (command == CMD_SET_MASQUERADE)
 	    command |= CMD_MASQUERADE;
@@ -1247,6 +1297,8 @@ main(int argc, char *argv[])
 		fw.ipfw.fw_tosxor = 0x00;
 	}
 
+	check_tos(fw.ipfw.fw_tosand, fw.ipfw.fw_tosxor, warnings);
+
 	if (options & OPT_JUMP)
 		strcpy(fw.label, jumpto);
 
@@ -1294,7 +1346,8 @@ main(int argc, char *argv[])
 		ret = list_entries(chain, command&CMD_ZERO,
 				   options&OPT_VERBOSE,
 				   options&OPT_NUMERIC,
-				   options&OPT_EXPANDED);
+				   options&OPT_EXPANDED,
+				   options&OPT_LINENUMBERS);
 		break;
 	case CMD_FLUSH:
 		ret = flush_entries(chain, options&OPT_VERBOSE);
@@ -1639,10 +1692,14 @@ parse_outputsize(char *outputstring, __u16 *outputsize)
 }
 
 static unsigned short
-parse_interface(const char *ifstring, char *vianame)
+parse_interface(const char *ifstring, char *vianame, int warnings)
 {
+	unsigned int i;
+	int ret;
 	if (strlen(ifstring) > IFNAMSIZ)
-		exit_error(PARAMETER_PROBLEM, "interface name `%s' must be shorter than IFNAMSIZ (%i)",
+		exit_error(PARAMETER_PROBLEM,
+			   "interface name `%s' must be shorter than"
+			   " IFNAMSIZ (%i)",
 			   ifstring, IFNAMSIZ);
 
 	strncpy(vianame, ifstring, IFNAMSIZ);
@@ -1650,9 +1707,23 @@ parse_interface(const char *ifstring, char *vianame)
 		return IP_FW_F_WILDIF;
 	else if(vianame[strlen(ifstring)-1] == '+') {
 		vianame[strlen(ifstring)-1] = '\0';
-		return IP_FW_F_WILDIF;
+		ret = IP_FW_F_WILDIF;
 	}
-	else return 0;
+	else ret = 0;
+
+	/* Warn if there are non alpha-numerics in interface name. */
+	if (warnings) {
+		for (i = 0; vianame[i]; i++) {
+			if (!isalnum(vianame[i])) {
+				printf("Warning: wierd character in interface"
+				       " `%s' (No aliases, :, ! or *).\n",
+				       vianame);
+				break;
+			}
+		}
+	}
+
+	return ret;
 }
 
 static __u16
@@ -1757,15 +1828,15 @@ host_to_addr(const char *name, unsigned int *naddr)
 }
 
 static char *
-addr_to_host(struct in_addr *addr)
+addr_to_host(const struct in_addr *addr)
 {
 	struct hostent *host;
 
-	if ((host = gethostbyaddr((char *) addr,
+	if ((host = gethostbyaddr((const char *) addr,
 			sizeof(struct in_addr), AF_INET)) != NULL)
-		return (char *) host->h_name;
+		return host->h_name;
 	else
-		return (char *) NULL;
+		return NULL;
 }
 
 static struct in_addr *
@@ -1784,18 +1855,18 @@ network_to_addr(const char *name)
 }
 
 static char *
-addr_to_network(struct in_addr *addr)
+addr_to_network(const struct in_addr *addr)
 {
 	struct netent *net;
 
 	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
-		return (char *) net->n_name;
+		return net->n_name;
 	else
-		return (char *) NULL;
+		return NULL;
 }
 
 static char *
-addr_to_anyname(struct in_addr *addr)
+addr_to_anyname(const struct in_addr *addr)
 {
 	char *name;
 
@@ -1844,18 +1915,18 @@ dotted_to_addr(const char *dotted)
 }
 
 static char *
-addr_to_dotted(struct in_addr *addrp)
+addr_to_dotted(const struct in_addr *addrp)
 {
 	static char buf[20];
-	unsigned char *bytep;
+	const unsigned char *bytep;
 
-	bytep = (unsigned char *) &(addrp->s_addr);
+	bytep = (const unsigned char *) &(addrp->s_addr);
 	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
 	return buf;
 }
 
 static char *
-mask_to_dotted(struct in_addr *mask)
+mask_to_dotted(const struct in_addr *mask)
 {
 	int i;
 	static char buf[20];
@@ -1982,7 +2053,7 @@ print_port(FILE *fp, int numeric, __u16 port, __u16 proto)
 
 static void
 print_firewall(FILE *fp,
-	       struct ip_fwuser *fw,
+	       const struct ip_fwuser *fw, int num,
 	       __u64 bytes, __u64 packets,
 	       int format)
 {
@@ -1994,6 +2065,8 @@ print_firewall(FILE *fp,
 #define FMT(tab,notab) ((format & FMT_NOTABLE) ? notab : tab)
 
 	if (format & FMT_HEADER) {
+		if (format & FMT_LINENUMBERS)
+			fprintf(fp, FMT("%-4s ", "%s "), "num");
 		if (!(format & FMT_NOCOUNTS)) {
 			if (format & FMT_KILOMEGAGIGA) {
 				fprintf(fp, FMT("%5s ","%s "), "pkts");
@@ -2022,6 +2095,9 @@ print_firewall(FILE *fp,
 		fputs(" ports\n", fp);
 	}
 
+	if (format & FMT_LINENUMBERS)
+		fprintf(fp, FMT("%-4d ", "%d "), num);
+
 	if (!(format & FMT_NOCOUNTS)) {
 		cnt = packets;
 		if (format & FMT_KILOMEGAGIGA) {
@@ -2066,7 +2142,8 @@ print_firewall(FILE *fp,
 	}
 
 	if (!(format & FMT_NOTARGET))
-		fprintf(fp, FMT("%-9s ", "%s "), fw->label);
+		fprintf(fp, FMT("%-9s ", "%s "), 
+			fw->label[0] ? fw->label : "-");
 
 	fputc(fw->ipfw.fw_invflg & IP_FW_INV_PROTO ? '!' : ' ', fp);
 	{
@@ -2099,14 +2176,11 @@ print_firewall(FILE *fp,
 
 	if (format & FMT_VIA) {
 		fputc(fw->ipfw.fw_invflg & IP_FW_INV_VIA ? '!' : ' ', fp);
-		if (fw->ipfw.fw_flg & IP_FW_F_WILDIF
-		    && (fw->ipfw.fw_vianame)[0]) {
-			fw->ipfw.fw_vianame[strlen(fw->ipfw.fw_vianame)+1]='\0';
-			fw->ipfw.fw_vianame[strlen(fw->ipfw.fw_vianame)]='+';
-		}
+		strcpy(buf, fw->ipfw.fw_vianame);
+		if (fw->ipfw.fw_flg & IP_FW_F_WILDIF && buf[0])
+			strcat(buf, "+");
 		fprintf(fp, FMT("%-10.16s ","via %.16s "),
-			(fw->ipfw.fw_vianame)[0] ? fw->ipfw.fw_vianame :
-				((format & FMT_NUMERIC) ? "*" : "any"));
+			buf[0] ? buf : ((format & FMT_NUMERIC) ? "*" : "any"));
 	}
 
 	if (format & FMT_MARK) {
@@ -2143,16 +2217,19 @@ print_firewall(FILE *fp,
 		fprintf(fp, FMT("%-20s ","%s "), buf);
 	}
 
-	fputc(fw->ipfw.fw_invflg & IP_FW_INV_DSTIP ? '!' : ' ', fp);
 	if (fw->ipfw.fw_dmsk.s_addr == 0L && !(format & FMT_NUMERIC))
-		fprintf(fp, FMT("%-20s","-> %s"), "anywhere");
+		fprintf(fp, FMT("%s%-20s","-> %s%s"), 
+			fw->ipfw.fw_invflg & IP_FW_INV_DSTIP ? "!" : "",
+			"anywhere");
 	else {
 		if (format & FMT_NUMERIC)
 			sprintf(buf, "%s", addr_to_dotted(&(fw->ipfw.fw_dst)));
 		else
 			sprintf(buf, "%s", addr_to_anyname(&(fw->ipfw.fw_dst)));
 		strcat(buf, mask_to_dotted(&(fw->ipfw.fw_dmsk)));
-		fprintf(fp, FMT("%-20s","-> %s"), buf);
+		fprintf(fp, FMT("%s%-20s","-> %s%s"), 
+			fw->ipfw.fw_invflg & IP_FW_INV_DSTIP ? "!" : "",
+			buf);
 	}
 
 	if (format & FMT_NOTABLE)
@@ -2221,7 +2298,7 @@ print_firewall(FILE *fp,
 	}
 
 	if (strcmp(fw->label, IP_FW_LABEL_REDIRECT) == 0) {
-		fputs(FMT(" => ", ""), fp);
+		fputs(" => ", fp);
 		if (fw->ipfw.fw_redirpt == 0)
 			fputs((format & FMT_NUMERIC) ? " *" : " any", fp);
 		else
@@ -2321,21 +2398,23 @@ read_masqinfo(FILE *fp, struct masq *mslist, int nmslist)
 }
 
 static void
-set_option(unsigned int *options, unsigned int option, __u16 *invflg)
+set_option(unsigned int *options, unsigned int option, __u16 *invflg,
+	   int *invert)
 {
 	if (*options & option)
 		exit_error(PARAMETER_PROBLEM, "multiple -%c flags not allowed\n",
 			   opt2char(option));
 	else *options |= option;
 
-	if (invert) {
+	if (*invert) {
 		unsigned int i;
 		for (i = 0; 1 << i != option; i++);
 		if (!inverse_for_options[i])
-			exit_error(PARAMETER_PROBLEM, "cannot have ! before -%c\n",
+			exit_error(PARAMETER_PROBLEM, 
+				   "cannot have ! before -%c\n",
 				   opt2char(option));
 		*invflg |= inverse_for_options[i];
-		invert = FALSE;
+		*invert = FALSE;
 	}
 }
 
@@ -2382,6 +2461,29 @@ fw_realloc(void *ptr, size_t size)
 	return p;
 }
 
+static void 
+warn_ipforward(void)
+{
+	FILE *fp = fopen("/proc/sys/net/ipv4/ip_forward", "r");
+
+	if (fp) {
+		int forward;
+	
+		if (fscanf(fp, "%d", &forward) && forward == 0) {
+			/* If inputting from a terminal, warn them. */
+			struct termios term;
+
+			if (tcgetattr(STDIN_FILENO, &term) == 0)
+				printf("Warning: you must enable IP forwarding"
+				       " for packets to be forwarded at all:\n"
+				       " Use `"
+				       "echo 1 > /proc/sys/net/ipv4/ip_forward"
+				       "'\n");
+		}
+		fclose(fp);
+	}
+}
+
 static void
 exit_error(enum exittype status, char *msg, ...)
 {
@@ -2411,20 +2513,14 @@ exit_printicmphelp()
 {
 	unsigned int i;
 	printf("%s\n\n"
-	       "Valid ICMP Types:",
+	       "Valid ICMP Types:\nType Code Description",
 	       package_version);
 
 	for (i = 0; i < sizeof(icmp_codes)/sizeof(struct icmp_names); i++) {
-		if (i && icmp_codes[i].type == icmp_codes[i-1].type) {
-			if (icmp_codes[i].code_min == icmp_codes[i-1].code_min
-			    && (icmp_codes[i].code_max
-				== icmp_codes[i-1].code_max))
-				printf(" (%s)", icmp_codes[i].name);
-			else
-				printf("\n   %s", icmp_codes[i].name);
-		}
-		else
-			printf("\n%s", icmp_codes[i].name);
+		if (icmp_codes[i].columns[0])
+			printf("\n%s %s",
+			       icmp_codes[i].columns, icmp_codes[i].name);
+		else printf(" (%s)", icmp_codes[i].name);
 	}
 	printf("\n");
 	exit(0);
@@ -2447,7 +2543,7 @@ exit_printhelp()
 	printf(
 "Commands:\n"
 "Either long or short options are allowed.\n"
-"  --add     -A chain		Append to chain\n"
+"  --append  -A chain		Append to chain\n"
 "  --delete  -D chain		Delete matching rule from chain\n"
 "  --delete  -D chain rulenum\n"
 "				Delete rule rulenum (1 = first) from chain\n"
@@ -2459,17 +2555,20 @@ exit_printhelp()
 "  --flush   -F [chain]		Delete all rules in  chain or all chains\n"
 "  --zero    -Z [chain]		Zero counters in chain or all chains\n"
 "  --check   -C chain		Test this packet on chain\n"
-"  --new     -N chain		Create a new user-defined chain\n"
-"  --delete  -X chain		Delete a user-defined chain\n"
+"  --new-chain\n"
+"            -N chain		Create a new user-defined chain\n"
+"  --delete-chain\n"
+"            -X chain		Delete a user-defined chain\n"
 "  --policy  -P chain target\n"
 "				Change policy on chain to target\n"
-"  --masquerade -M -L		List current masqerading connections\n"
-"  --masquerade -M -S tcp tcpfin udp\n"
+"  --masquerading\n"
+"            -M -L		List current masquerading connections\n"
+"  --set     -M -S tcp tcpfin udp\n"
 "				Set masquerading timeout values\n\n"
 
 "Options:\n"
 "  --bidirectional -b		insert two rules: one with -s & -d reversed\n"
-"  --proto	-p [!] proto	protocol: by number or name, eg. `tcp'\n"
+"  --protocol	-p [!] proto	protocol: by number or name, eg. `tcp'\n"
 "  --source	-s [!] address[/mask] [!] [port[:port]]\n"
 "				source specification\n"
 "  --source-port [!] [port[:port]]\n"
@@ -2490,6 +2589,8 @@ exit_printhelp()
 "  --TOS		-t and xor	and/xor masks for TOS field\n"
 "  --verbose	-v		verbose mode\n"
 "  --exact	-x		expand numbers (display exact values)\n"
+"  --line-numbers		list line numbers beside each rule\n"
+"  --no-warnings		Disable all warnings\n"
 "[!] --fragment	-f		match second or further fragments only\n"
 "[!] --syn	-y		match TCP packets only when SYN set\n"
 "[!] --version	-V		print package version.\n");
@@ -2561,7 +2662,7 @@ count_tos_bits(unsigned char bitpattern)
 }
 
 static void
-check_tos(unsigned char tosand, unsigned char tosxor)
+check_tos(unsigned char tosand, unsigned char tosxor, int warning)
 {
 	unsigned int i;
 
@@ -2580,16 +2681,17 @@ check_tos(unsigned char tosand, unsigned char tosxor)
 			   "(RFC1349)\n");
 
 	for (i = 0; i < sizeof(TOS_values)/sizeof(struct TOS_value); i++)
-		if (count_tos_bits((TOS_values[i].TOS & tosand) ^ tosxor) > 1)
+		if (count_tos_bits((TOS_values[i].TOS & tosand) ^ tosxor) > 1
+		    && warning)
 			printf("Warning: TOS manipulation may set multiple "
 			       "TOS bits for %s TOS (0x%02x)\n",
 			       TOS_values[i].name, TOS_values[i].TOS);
 }
 
 static void
-add_command(int *cmd, const int newcmd, const int othercmds)
+add_command(int *cmd, const int newcmd, const int othercmds, int *invert)
 {
-	if (invert)
+	if (*invert)
 		exit_error(PARAMETER_PROBLEM, "unexpected ! flag");
 	if (*cmd & (~othercmds))
 		exit_error(PARAMETER_PROBLEM, "Can't use -%c with -%c\n",
